import cadquery as cq
import math
from cq_server.ui import ui, show_object

# Bitting (only five values work)
# Buffo demo lock: 7-5-1-1-(2 oder 3)
#bitting = [7, 5, 1, 1, 2]

# C83 half cylinder
bitting = [2, 4, 7, 2, 5]

# Dimensions for the key
# Blade
key_blade_length = 27
key_blade_height = 8.25
key_blade_thickness = 2.2
# Bow
key_bow_height = 10.2
key_bow_thickness = key_blade_thickness
key_bow_length = 7
# Handle
key_handle_height = 14 
key_handle_thickness = key_blade_thickness
key_handle_length = 10 
key_handle_offset_from_center = 3


### 
# Define the key cutter and bitting for ABUS C83
###
# What angle does the cutter have?
cutter_angle_degree = 112
cutter_tooth_x = 10 # must be greater than 0-cut
cutter_tooth_y = 6.35
# the flat tip of the cutter
cutter_tip_y = 1.12
#bitting_cut_angle_degree = 90
# Distances from bow:
bitting_bow_distance = {
    1: 4.42,
    2: 8.42,
    3: 12.42,
    4: 16.42,
    5: 20.45,
}
# Bitting depth (from opposite side of bow)
bitting_cut_depth = {
    9: 4.45,
    8: 4.66,
    7: 5.1,
    6: 5.53,
    5: 5.96,
    4: 6.40,
    3: 6.83,
    2: 7.26,
    1: 7.70,
}

# Helper to create a cutter for a single cut, positioned correctly 
# in relation to the blade
def create_cutter_tooth(no_of_cut, bitting):
    # Create the desired cutting angle
    delta_x = ((cutter_tooth_y-cutter_tip_y)/2) / math.sin(cutter_angle_degree)
    cutter_tooth_outline = [
        (-cutter_tooth_x, -cutter_tooth_y/2),
        (-cutter_tooth_x, cutter_tooth_y/2),
        #(-cutter_tooth_y/2, cutter_tooth_y/2),
        (delta_x, +cutter_tooth_y/2),
        (0, +cutter_tip_y/2),
        (0, 0), # this is the center of the tip
        (0, -cutter_tip_y/2),
        (delta_x, -cutter_tooth_y/2),
        #(-cutter_tooth_y/2, -cutter_tooth_y/2),
        (-cutter_tooth_x,-cutter_tooth_y/2)
    ]
    cut_depth = key_blade_height-bitting_cut_depth[bitting]
    cut_distance = key_blade_length-bitting_bow_distance[no_of_cut]
    cutter_tooth = cq.Workplane("YZ").workplane(offset=-2)\
        .polyline(cutter_tooth_outline).close().extrude(6)\
        .translate((
            0, 
            cut_depth, 
            cut_distance
        ))
    print(f"- cut {no_of_cut} ({cut_distance:04.2f} mm): bitting {bitting} (remaining blade thickness: {(bitting_cut_depth[bitting]):04.2f} mm)")
    return cutter_tooth


###
# Create the key
###
# Create the blade profile as a 2D object
profile = (
    # see https://cadquery.readthedocs.io/en/latest/primer.html#fluent-api-direct-api
    cq.importers.importDXF("profile/AB-AB1.dxf") \
        .wires().toPending() \
        #.extrude(key_blade_length) \
        .val() # Go from fluent to direct API: return a shape,
)
# profile is now a 2D outline.
profile_bb = profile.BoundingBox()
print(f"X_len: {profile_bb.xlen}, ratio: {profile_bb.xlen/key_blade_thickness}")
print(f"Y_len: {profile_bb.ylen}, ratio: {profile_bb.ylen/key_blade_height}")
# Scale to the correct dimensions.
profile = profile.scale(1/(profile_bb.ylen/key_blade_height))
# Back to fluent API: create the blade, based on the profile
blade = cq.Workplane(obj = profile) \
    .wires().toPending().extrude(key_blade_length)
# stencil to cut tip of blade
blade_tip_cutter = [
    (5, 0),
    (9, 6),
    (9, 0),
    (5, 0),
    (0, 5),
    (0, 0),
    (5, 0),
]
blade = blade.faces(">Y").cut(
    cq.Workplane("YZ").polyline(blade_tip_cutter).close().extrude(10)
)
# Add the bow: shoulder stop for keyway
blade = blade.faces(">Z").workplane()\
    .center(key_handle_thickness/2, key_handle_offset_from_center)\
    .rect(key_bow_thickness, key_bow_height)\
    .extrude(key_bow_length)
# Add lofted cylinder as a handle
blade = blade.faces(">Z").workplane()\
    .rect(key_bow_thickness, key_bow_height)\
    .workplane(offset=6)\
    .circle(key_handle_height/2)\
    .loft(combine=True)
blade = blade.faces(">Z").workplane() \
    .circle(key_handle_height/2)\
    .extrude(10)    

# create the cutter geometry for the complete bitting value
cutter = []
for idx, b in enumerate(bitting):
    tooth = create_cutter_tooth(idx+1, b)
    if cutter == []:
        # first tooth of cutter, no union to perform
        cutter = tooth
    else:
        cutter = cutter.union(tooth)

# Now: Create the final key. Subtract the cutter from the key's blade.
key = blade.cut(cutter)

# Notched key: An experiment. Add notches so that an external turning tool
# can be used to turn the keyway.
turner_cutter_length = 8
turner_cutter_height = 10
turner_cutter_thickness = 3*key_blade_thickness
# two cutters
turner_cutter1 = cq.Workplane("YZ").workplane()\
    .box(turner_cutter_height, turner_cutter_length, turner_cutter_thickness)\
    .translate((0, -4, key_blade_length+2))\
    .edges("|X").fillet(1)
turner_cutter2 = cq.Workplane("YZ").workplane()\
    .box(turner_cutter_height, turner_cutter_length, turner_cutter_thickness)\
    .translate((0, key_blade_height+4, key_blade_length+2))\
    .edges("|X").fillet(1)
# Create the notched key by subtracting the cutters from the key
notched_key = key.cut(turner_cutter1).cut(turner_cutter2)

# Render the solids
show_object(key)
#show_object(cutter)
#show_object(turner_cutaway1)
#show_object(turner_cutaway2)
#show_object(notched_key)

# Write STL for key and notched key
bitting_string_list=[str(i) for i in bitting]
bitting_string = "-".join(bitting_string_list)
key_filename = f"ABUS-C83-{bitting_string}.stl"
notched_key_filename = f"ABUS-C83-{bitting_string}-notched.stl"
cq.exporters.export(key, key_filename)
cq.exporters.export(notched_key, notched_key_filename)